@@ -125,10 +125,19 @@ impl connection::Trait for TestConnection {
         Ok(())
     }
 
+    fn on_pending_ack_ranges(
+        &mut self,
+        _timestamp: Timestamp,
+        _subscriber: &mut <Self::Config as endpoint::Config>::EventSubscriber,
+    ) -> Result<(), connection::Error> {
+        Ok(())
+    }
+
     fn on_wakeup(
         &mut self,
         _timestamp: Timestamp,
         _subscriber: &mut <Self::Config as endpoint::Config>::EventSubscriber,
+        _datagram: &mut <Self::Config as endpoint::Config>::DatagramEndpoint,
     ) -> Result<(), connection::Error> {
         Ok(())
     }
@@ -141,6 +150,7 @@ impl connection::Trait for TestConnection {
         _random_generator: &mut <Self::Config as endpoint::Config>::RandomGenerator,
         _subscriber: &mut <Self::Config as endpoint::Config>::EventSubscriber,
         _packet_interceptor: &mut <Self::Config as endpoint::Config>::PacketInterceptor,
+        _datagram_endpoint: &mut <Self::Config as endpoint::Config>::DatagramEndpoint,
     ) -> Result<(), ProcessingError> {
         Ok(())
     }
@@ -154,6 +164,7 @@ impl connection::Trait for TestConnection {
         _random_generator: &mut <Self::Config as endpoint::Config>::RandomGenerator,
         _subscriber: &mut <Self::Config as endpoint::Config>::EventSubscriber,
         _packet_interceptor: &mut <Self::Config as endpoint::Config>::PacketInterceptor,
+        _datagram_endpoint: &mut <Self::Config as endpoint::Config>::DatagramEndpoint,
     ) -> Result<(), ProcessingError> {
         Ok(())
     }
@@ -167,6 +178,7 @@ impl connection::Trait for TestConnection {
         _random_generator: &mut <Self::Config as endpoint::Config>::RandomGenerator,
         _subscriber: &mut <Self::Config as endpoint::Config>::EventSubscriber,
         _packet_interceptor: &mut <Self::Config as endpoint::Config>::PacketInterceptor,
+        _datagram_endpoint: &mut <Self::Config as endpoint::Config>::DatagramEndpoint,
     ) -> Result<(), ProcessingError> {
         Ok(())
     }
